% -------------------------------------------------------------------------
%       Topology-optimization based design of multi-degree-of-freedom 
%       compliant mechanisms (mechanisms with multiple pseudo-mobility)
%
%         Stephanie Seltmann, Lucio Flavio Campanile, Alexander Hasse
%            Journal of Intelligent Material Systems and Structures
% -------------------------------------------------------------------------

% *************************** Disclaimer *********************************
% The authors reserve all rights but do not guarantee that the code is
% free from errors. In addition, they are in no case liable for any event 
% caused by the use of the program.  
% ************************************************************************

% the numbers in brackets correspond to equations from the paper
clear all
%%%%%%%%%%%%%%%%%%%%%%Input of the required parameters%%%%%%%%%%%%%%%%%%%%%
%parametrization
bb=40;                  %width of design space [mm]
h=30;                   %height of design space [mm]
numb=16;                %number of elements on width
numh=12;                %number of elements on height
%material and dimensions of beam elements
width=10;               %width beam elements [mm]
thick=2;                %thickness of beam elements [mm]
E=210000;               %Young's modulus [N/mm^2]
volumefraction=0.8;     %distributable volume based on initial volume
%boundary conditions
%%%%%%%%%%%%%%%%%%
%node arrangement%
%       3 4      %
%       1 2      %
%%%%%%%%%%%%%%%%%%
fixeddofs=1:51;         %fixed degrees of freedom
%subspace of desired deformations, each line is a desired deformation mode,
%vectors of desired deformation modes should be orthogonal to each other,
%please note that the program is set for 1 to 3 desired deformation modes
phi=[1 0 1 0;...
    0 -1 0 1];
modedofs=[613 614 661 662]; %active DoFs
%parameters for optimization
numM=2;                 %number of modes for stabilization according to (29)
v=0.001;                %nu (36)
numIte=1500;            %number of iterations
mu=1000;                %upper limit to the primary stiffness mu

%%%%%%%%%%%%%%%%%%%%%%%%%calculations%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%normalization of phi according to (15)
phi=(1./diag(sqrt(phi*phi'))*ones(1,size(phi,2)).*phi)';

%allocation matrix and node coordinates
[ZUO,PuKo]=BeamTruss(h,bb,numb,numh);
beams=length(ZUO);
%calculation of stiffness matrix
[Standardmatrix]=Standardmatrices(ZUO,PuKo,width,thick,E);
%subdivision of degrees of freedom
alldofs = 1:length(PuKo)*3;             %all degrees of freedom
freedofs = setdiff(alldofs,fixeddofs);  %free degrees of freedom
modefree  = setdiff(freedofs,modedofs); %passive DoFs
%limits according to (19)
xmin=0.00000001*ones(beams,1);
xmax=1*ones(beams,1);
%beam degrees of freedom matrix
BEAMS=zeros(beams,6);
for i=1:beams
    BEAMS(i,:)=[(3*(ZUO(i,2))-2) (3*(ZUO(i,2))-2)+1 (3*(ZUO(i,2))-2)+2 ...
        (3*(ZUO(i,3))-2) (3*(ZUO(i,3))-2)+1 (3*(ZUO(i,3))-2)+2];
end
iK = reshape(kron(BEAMS,ones(6,1))',36*beams,1);
jK = reshape(kron(BEAMS,ones(1,6))',36*beams,1);
standard=cell2mat(Standardmatrix);
d3=zeros(6,6,beams);
for i=1:beams
    d3(:,:,i)=Standardmatrix{1,i};
end
NumMode=length(phi(1,:));

%start vector 
x0=ones(length(xmin),1);
fvalg=zeros(numIte,1);

%optimization options
options = optimoptions(@linprog,'Algorithm','dual-simplex',...
    'MaxIter',3000,'Preprocess','none','Tolcon',1e-3,'Display','none');

%%%%%%%%%%%%%%%%%%%%%global iteration procedure%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for i=1:numIte
    %%%%%%%%%%%%%calculation objective function and restrictions%%%%%%%%%%%
    %objective function (26), restrictions (27), (28) and (29)
    [A,f,Aeq,beq]=Displacement_distribution(x0,alldofs,...
    phi,modedofs,numM,BEAMS,modefree,iK,jK,standard,d3,beams,NumMode);
    b(1:NumMode)=mu;
    b(NumMode+1:length(A(:,1)))=0;
    %volume_constraint (30)
    A(length(A(:,1))+1,:)=ones(1,beams);
    b(length(A(:,1)))=beams*volumefraction;
    %%%%%%%%%%%sub-problem 2: updating the design variables%%%%%%%%%%%%%%%%
    [x,fval,exitflag,output]=linprog(f,A,b,Aeq,beq,xmin,xmax,...
        options);
    if exitflag==-2 || exitflag==0
        x=x0;
        disp(['no solution can be found with these input parameters and'...
        ' this starting vector, please try again']);
        break;
    end
    %convergence history
    fvalg(i)=fval;
    
    %%%%%%%%%%%stabilization of the global iteration procedure%%%%%%%%%%%%%
    % stabilization (32)
    x0=x;
    xmax=x0+v;
    xmax(xmax>1)=1;
    xmin=x0-v;
    xmin(xmin<0.00000001)=0.00000001;
    disp(['Iteration: ' num2str(i)]);
end

%%%%%%%%plot the result and eigenvalue analysis according to (4)%%%%%%%%%%%
[chi,lambda]=Plotdef(x,ZUO,PuKo,freedofs,alldofs,h,bb,modedofs,thick,...
    beams,modefree,standard,iK,jK);
%eigenvalue analysis
[lambdasort,ind]=sort(diag(lambda));        %eigenvalues
chisort=chi(:,ind);                         %eigenvectors
S=lambdasort(NumMode+1)/lambdasort(NumMode);%selectivity according to (11)